from ..ibuiltins.is_empty import is_empty


def irange(start, finish, step=1):
    """
    Meningkatkan fungsi range() dari python untuk pengulangan menggunakan huruf

    ```python
    print(irange('a', 'c'))
    print(irange('z', 'a', 10))
    print(list(irange('a', 'z', 10)))
    print(list(irange(1, '7')))
    print(list(irange(10, 5)))
    ```
    """

    def casting_class():
        start_int = isinstance(start, int)
        finish_int = isinstance(finish, int)
        start_str = isinstance(start, str)
        finish_str = isinstance(finish, str)
        start_numeric = start.isnumeric() if start_str else False
        finish_numeric = finish.isnumeric() if finish_str else False

        if start_numeric and finish_numeric:
            # irange("1", "5")
            return (int, str)

        if (start_numeric or start_int) and (finish_numeric or finish_int):
            # irange("1", "5")
            # irange("1", 5)
            # irange(1, "5")
            # irange(1, 5)
            return (int, int)

        if start_str and finish_str:
            # irange("a", "z")
            # irange("p", "g")
            return (ord, chr)

        """
        kedua if dibawah ini sudah bisa berjalan secara logika, tetapi
        perlu dimanipulasi supaya variabel start dan finish bisa diubah.
        """
        # irange(1, 'a') -> irange('1', 'a')
        # irange(1, '5') -> irange(1, 5)
        # irange('1', 5) -> irange(1, 5)
        # irange('a', 5) -> irange('a', '5')
        #
        # if start_str and finish_int:
        #     # irange("a", 5) -> irange("a", "5")
        #     finish = str(finish)
        #     return (ord, chr)
        #
        # if start_int and finish_str:
        #     # irange(1, "g") -> irange("1", "g")
        #     start = str(start)
        #     return (ord, chr)

        raise Exception(f"[{start} - {finish}] tidak dapat diidentifikasi kesamaannya")

    counter_class, converter_class = casting_class()
    start = counter_class(start)
    finish = counter_class(finish)

    step = 1 if is_empty(step) else int(step)

    faktor = 1 if finish > start else -1
    step *= faktor
    finish += faktor

    for i in range(start, finish, step):
        yield converter_class(i)


if __name__ == "__main__":
    from .iargv import iargv
    from .iprint import iprint

    if (s := iargv(1)) and (f := iargv(2)):
        iprint(irange(s, t))



def dict_first(d: dict, remove=False):
    """
    Mengambil nilai (key, value) pertama dari dictionary dalam bentuk tuple.

    ```python
    d = {
        "key2": "value2",
        "key3": "value3",
        "key1": "value1",
    }
    print(dict_first(d, remove=True))
    print(dict_first(d))
    ```
    """
    for k in d:
        return (k, d.pop(k) if remove else d[k])



import random


def random_bool():
    """
    Menghasilkan nilai random True atau False.
    Fungsi ini merupakan fungsi tercepat untuk mendapatkan random bool.
    Fungsi ini sangat cepat, tetapi pemanggilan fungsi ini membutuhkan
    overhead yg besar.

    ```python
    print(random_bool())
    ```
    """
    return bool(random.getrandbits(1))


if __name__ == "__main__":
    print(random_bool())



from ..iconsole.console_run import console_run


def github_user(email=None, name=None):
    """
    Menyimpan email dan nama user secara global sehingga tidak perlu
    menginput nya setiap saat.

    ```py
    github_user('my@emil.com', 'MyName')
    ```
    """
    if email:
        console_run(
            "Update Github User Email", f"git config --global user.email {email}"
        )
    if name:
        console_run("Update Github User Name", f"git config --global user.name {name}")


if __name__ == "__main__":
    from ..ifunctions.iargv import iargv

    github_user(iargv(1), iargv(2))



from ..ibuiltins.is_iterable import is_iterable
from ..ibuiltins.WINDOWS import WINDOWS
from .print_to_last_line import print_to_last_line
from .text_colorize import text_colorize


def choices(iterator, title=None, prompt="", default=None):
    """
    Memudahkan dalam membuat pilihan untuk user dalam tampilan console

    ```py
    a = choices("ini hanya satu pilihan")
    b = choices(
        {
            "sedan": "audi",
            "suv": "volvo",
            "truck": "tesla",
        },
        title="Car Model",
        prompt="Pilih Mobil : ",
    )
    c = choices(
        iscandir(recursive=False),
        title="List File dan Folder",
        prompt="Pilih File atau Folder : ",
    )
    ```
    """

    def build_iterator(iterator):
        if not is_iterable(iterator):
            iterator = [iterator]
        if not isinstance(iterator, dict):
            iterator = dict(enumerate(iterator))
        return iterator

    def print_choices(dictionary, title=None):
        if title:
            print(title)
        for k, v in dictionary.items():
            print(f"[{k}] {v}")

    def input_choices(prompt):
        i = input(prompt)
        if WINDOWS:
            i = i.decode()
        if i.isdigit():
            i = int(i)
        return i

    def return_choices(iterator, key, on_error):
        try:
            result = iterator[key]
        except Exception:
            result = on_error
        return result

    d = build_iterator(iterator)
    print_choices(d, title)
    key = input_choices(prompt)
    result = return_choices(d, key, default)
    r = text_colorize(result)
    print_to_last_line(f"{prompt}{key} [{r}]")
    return result



import sys


def iargv(key: int, cast=None, on_error=None):
    try:
        v = sys.argv[key]
        if cast:
            return cast(v)
        else:
            return v
    except Exception:
        return on_error




def is_iterable(var, str_is_iterable=False):
    """
    Mengecek apakah suatu variabel bisa dilakukan forloop atau tidak

    ```python
    s = 'ini string'
    print(is_iterable(s))

    l = [12,21,2,1]
    print(is_iterable(l))

    r = range(100)
    print(is_iterable(r))

    d = {'a':1, 'b':2}
    print(is_iterable(d.values()))
    ```
    """

    """ Metode #1 """
    # TYPE_NONE = type(None)
    # TYPE_GENERATOR = type(i for i in [])
    # TYPE_RANGE = type(range(0))
    # TYPE_DICT_VALUES = type(dict().values())
    # it = (list, tuple, set, dict)
    # it += (TYPE_GENERATOR, TYPE_RANGE, TYPE_DICT_VALUES)
    # return isinstance(var, it)

    """ Metode #2 """
    if isinstance(var, str) and not str_is_iterable:
        return False
    # return isinstance(var, collections.abc.Iterable)

    """ Metode #3 """
    try:
        iter(var)
        return True
    except Exception:
        return False



import re

from ..ifunctions.irange import irange

# batchmaker()
__batchmaker__regex_pattern__ = r"{(?:[^a-zA-Z0-9{}]+)?([a-zA-Z]|\d+)[^a-zA-Z0-9{}]+([a-zA-Z]|\d+)(?:[^a-zA-Z0-9{}]+(\d+))?(?:[^a-zA-Z0-9{}]+)?}"
__batchmaker__regex_compile__ = re.compile(__batchmaker__regex_pattern__)


def batchmaker(pattern: str):
    """
    Alat Bantu untuk membuat teks yang berulang.
    Gunakan `{[start][separator][finish]([separator][step])}`.
    ```
    [start] dan [finish]    -> bisa berupa huruf maupun angka
    ([separator][step])     -> bersifat optional
    [separator]             -> selain huruf dan angka
    [step]                  -> berupa angka positif
    ```

    ```python
    s = "Urutan {1/6/3} dan {10:9} dan {j k} dan {Z - A - 15} saja."
    print(batchmaker(s))
    print(list(batchmaker(s)))
    ```
    """
    s = __batchmaker__regex_compile__.search(pattern)
    if s is None:
        yield pattern
        return

    find = s.group()
    start, finish, step = s.groups()

    for i in irange(start, finish, step):
        r = pattern.replace(find, i, 1)
        yield from batchmaker(r)


if __name__ == "__main__":
    from ..ifunctions.iargv import iargv
    from ..ifunctions.iprint import iprint

    if i := iargv(1):
        iprint(batchmaker(i))



from .ibuiltins.ComparePerformance import ComparePerformance
from .ibuiltins.LINUX import LINUX
from .ibuiltins.RunParallel import RunParallel
from .ibuiltins.WINDOWS import WINDOWS
from .ibuiltins.avg import avg
from .ibuiltins.basename import basename
from .ibuiltins.chunk_array import chunk_array
from .ibuiltins.create_folder import create_folder
from .ibuiltins.datetime_from_string import datetime_from_string
from .ibuiltins.datetime_now import datetime_now
from .ibuiltins.dict_first import dict_first
from .ibuiltins.dirname import dirname
from .ibuiltins.exit_if_empty import exit_if_empty
from .ibuiltins.filter_empty import filter_empty
from .ibuiltins.get_class_method import get_class_method
from .ibuiltins.get_filemtime import get_filemtime
from .ibuiltins.get_filesize import get_filesize
from .ibuiltins.is_empty import is_empty
from .ibuiltins.is_iterable import is_iterable
from .ibuiltins.is_valid_url import is_valid_url
from .ibuiltins.password_generator import password_generator
from .ibuiltins.random_bool import random_bool
from .ibuiltins.set_timeout import set_timeout
from .ibuiltins.sets_ordered import sets_ordered
from .ibuiltins.str_cmp import str_cmp
from .ibuiltins.to_str import to_str
from .iconsole.choices import choices
from .iconsole.console_run import console_run
from .iconsole.input_char import input_char
from .iconsole.log import log
from .iconsole.print_colorize import print_colorize
from .iconsole.print_dir import print_dir
from .iconsole.print_log import print_log
from .iconsole.print_to_last_line import print_to_last_line
from .iconsole.text_colorize import text_colorize
from .idjango.APIMixinView import APIMixinView
from .iengineering.PintUreg import PintUreg
from .iengineering.PintUregQuantity import PintUregQuantity
from .iengineering.batch_calculate import batch_calculate
from .iengineering.batchmaker import batchmaker
from .iengineering.calculate import calculate
from .iflow.auto_reload import auto_reload
from .iflow.github_pull import github_pull
from .iflow.github_push import github_push
from .iflow.github_user import github_user
from .iflow.pip_freeze_without_version import pip_freeze_without_version
from .iflow.poetry_publish import poetry_publish
from .iflow.poetry_update_version import poetry_update_version
from .ifunctions.iargv import iargv
from .ifunctions.idumps import idumps
from .ifunctions.idumps_html import idumps_html
from .ifunctions.ienv import ienv
from .ifunctions.iexec import iexec
from .ifunctions.ijoin import ijoin
from .ifunctions.iloads import iloads
from .ifunctions.iloads_html import iloads_html
from .ifunctions.iopen import iopen
from .ifunctions.iprint import iprint
from .ifunctions.irange import irange
from .ifunctions.ireplace import ireplace
from .ifunctions.iscandir import iscandir
from .ifunctions.isplit import isplit

print("Ga boleh Import pypipr.pypipr")

import sys
sys.exit(13)




import re


def isplit(text, separator="", include_separator=False):
    """
    Memecah text menjadi list berdasarkan separator.

    ```python
    t = '/ini/contoh/path/'
    print(isplit(t, separator='/'))
    ```
    """
    if include_separator:
        separator = f"({separator})"

    result = re.split(separator, text, flags=re.IGNORECASE | re.MULTILINE)

    return result



import os


def get_filemtime(filename):
    """
    Mengambil informasi last modification time file dalam nano seconds

    ```python
    print(get_filemtime(__file__))
    ```
    """
    return os.stat(filename).st_mtime_ns



import subprocess
import time

from ..ibuiltins.get_filemtime import get_filemtime
from ..iconsole.print_log import print_log


def auto_reload(filename):
    mtime = get_filemtime(filename)
    last_mtime = 0
    cmd = f"python {filename}"

    try:
        print_log("Start")
        while True:
            last_mtime = mtime
            subprocess.run(cmd, shell=True)
            while mtime == last_mtime:
                time.sleep(1)
                mtime = get_filemtime(filename)
            print_log("Reload")
    except KeyboardInterrupt:
        print_log("Stop")


if __name__ == "__main__":
    from pypipr.ifunctions.iargv import iargv

    if f := iargv(1):
        auto_reload(f)



import asyncio
import multiprocessing
import queue
import threading


class RunParallel:
    """
    Menjalankan program secara bersamaan.

    - `class RunParallel` didesain hanya untuk pemrosesan data saja.
    - Penggunaannya `class RunParallel` dengan cara membuat instance
      sub class beserta data yg akan diproses, kemudian panggil fungsi
      yg dipilih `run_asyncio / run_multi_threading / run_multi_processing`,
      kemudian dapatkan hasilnya.
    - `class RunParallel` tidak didesain untuk menyimpan data, karena
      setiap module terutama module `multiprocessing` tidak dapat mengakses
      data kelas dari proses yg berbeda.
    - Semua methods akan dijalankan secara paralel kecuali method dengan
      nama yg diawali underscore `_`
    - Method untuk multithreading/multiprocessing harus memiliki 2
      parameter, yaitu: `result: dict` dan `q: queue.Queue`. Parameter
      `result` digunakan untuk memberikan return value dari method, dan
      Parameter `q` digunakan untuk mengirim data antar proses.
    - Method untuk asyncio harus menggunakan keyword `async def`, dan
      untuk perpindahan antar kode menggunakan `await asyncio.sleep(0)`,
      dan keyword `return` untuk memberikan return value.
    - Return Value berupa dictionary dengan key adalah nama function,
      dan value adalah return value dari setiap fungsi
    - Menjalankan Multiprocessing harus berada dalam blok
      `if __name__ == "__main__":` karena area global pada program akan
      diproses lagi. Terutama pada sistem operasi windows.
    - `run_asyncio()` akan menjalankan kode dalam satu program, hanya
      saja alur program dapat berpindah-pindah menggunkan
      `await asyncio.sleep(0)`.
    - `run_multi_threading()` akan menjalankan program dalam satu CPU,
      hanya saja dalam thread yang berbeda. Walaupun tidak benar-benar
      berjalan secara bersamaan namun bisa meningkatkan kecepatan
      penyelesaian program, dan dapat saling mengakses resource antar
      program.  Akses resource antar program bisa secara langsung maupun
      menggunakan parameter yang sudah disediakan yaitu `result: dict`
      dan `q: queue.Queue`.
    - `run_multi_processing()` akan menjalankan program dengan beberapa
      CPU. Program akan dibuatkan environment sendiri yang terpisah dari
      program induk. Keuntungannya adalah program dapat benar-benar berjalan
      bersamaan, namun tidak dapat saling mengakses resource secara langsung.
      Akses resource menggunakan parameter yang sudah disediakan yaitu
      `result: dict` dan `q: queue.Queue`.

    ```python
    class ExampleRunParallel(RunParallel):
        z = "ini"

        def __init__(self) -> None:
            self.pop = random.randint(0, 100)

        def _set_property_here(self, v):
            self.prop = v

        def a(self, result: dict, q: queue.Queue):
            result["z"] = self.z
            result["pop"] = self.pop
            result["a"] = "a"
            q.put("from a 1")
            q.put("from a 2")

        def b(self, result: dict, q: queue.Queue):
            result["z"] = self.z
            result["pop"] = self.pop
            result["b"] = "b"
            result["q_get"] = q.get()

        def c(self, result: dict, q: queue.Queue):
            result["z"] = self.z
            result["pop"] = self.pop
            result["c"] = "c"
            result["q_get"] = q.get()

        async def d(self):
            print("hello")
            await asyncio.sleep(0)
            print("hello")

            result = {}
            result["z"] = self.z
            result["pop"] = self.pop
            result["d"] = "d"
            return result

        async def e(self):
            print("world")
            await asyncio.sleep(0)
            print("world")

            result = {}
            result["z"] = self.z
            result["pop"] = self.pop
            result["e"] = "e"
            return result

    if __name__ == "__main__":
        print(ExampleRunParallel().run_asyncio())
        print(ExampleRunParallel().run_multi_threading())
        print(ExampleRunParallel().run_multi_processing())
    ```
    """

    def get_all_instance_methods(self, coroutine):
        c = set(dir(__class__))
        t = (x for x in dir(self) if x not in c)
        return tuple(
            a
            for x in t
            if callable(a := getattr(self, x))
            and not x.startswith("_")
            and asyncio.iscoroutinefunction(a) == coroutine
        )

    def run_asyncio(self):
        m = self.get_all_instance_methods(coroutine=True)
        a = self.module_asyncio(*m)
        return self.dict_results(m, a)

    def run_multi_threading(self):
        m = self.get_all_instance_methods(coroutine=False)
        a = self.module_threading(*m)
        return self.dict_results(m, a)

    def run_multi_processing(self):
        m = self.get_all_instance_methods(coroutine=False)
        a = self.module_multiprocessing(*m)
        return self.dict_results(m, a)

    def dict_results(self, names, results):
        return dict(zip((x.__name__ for x in names), results))

    def module_asyncio(self, *args):
        async def main(*args):
            return await asyncio.gather(*(x() for x in args))

        return asyncio.run(main(*args))

    def module_threading(self, *args):
        a = tuple(dict() for _ in args)
        q = queue.Queue()
        r = tuple(
            threading.Thread(target=v, args=(a[i], q)) for i, v in enumerate(args)
        )
        for i in r:
            i.start()
        for i in r:
            i.join()
        return a

    def module_multiprocessing(self, *args):
        m = multiprocessing.Manager()
        q = m.Queue()
        a = tuple(m.dict() for _ in args)
        r = tuple(
            multiprocessing.Process(target=v, args=(a[i], q))
            for i, v in enumerate(args)
        )
        for i in r:
            i.start()
        for i in r:
            i.join()
        return (i.copy() for i in a)
